// bingo2 for PLB (C) 2024 Shimeji Sola
// A bingo client played by each of the attendees

// Global variables of the client, shared among the screens
class ClientEnvironment {
  constructor() {
    this.roomCode = "";
    this.cardID = "";
  }
}

var game = new ClientEnvironment();

/**********************************************************************/
// title screen of the client, room code is determined here
class TitleScreen extends Phaser.Scene {
  constructor( ...args ) {
    super({ key: 'title', ...args });
  }

  create() {
    this.roomCode = game.roomCode;

    // texts
    this.add.text(350, 150, 'Bingo 2 for PLB', env.defaultFont).setOrigin(0.5, 0);
    this.add.text(350, 250, 'Room Code', env.defaultFont).setOrigin(0.5, 0);
    this.roomStatusText = this.add.text(350, 1090, 'Enter Room Code', env.defaultFont).setOrigin(0.5, 0);
    let continueRoom = this.add.text(930, 1090, 'No Existing Cards', env.defaultFont).setOrigin(0.5, 0);
    let continueCard = this.add.text(930, 1160, 'Card ID: ---------', env.defaultFont).setOrigin(0.5, 0);

    // buttons
    let startFunc = this.checkRoom.bind(this);
    let enterNumberFunc = this.enterNumber.bind(this);
    this.startButton = new GameButton(this, "Start", 100, 850, 500, 200, false, startFunc);
    let continueButton = new GameButton(this, "Continue", 680, 850, 500, 200, false, startFunc);
    this.numberButton = [];
    for (let y = 0; y < 4; y++) {
      for (let x = 0; x < 3; x++) {
        if (y != 3 || x != 0) {
          let n = "123456789 0<"[x + y * 3];
          let newButton = new GameButton(this, n, 680 + x * 180, 80 + y * 180, 150, 150, true, enterNumberFunc);
          this.numberButton.push(this.add.existing(newButton));
        }
      }
    }
    if (this.roomCode.length != 4)
      this.startButton.disable();
    continueButton.disable();
    if (this.roomCode.length == 0)
      this.numberButton[10].disable();
    this.add.existing(this.startButton);
    this.add.existing(continueButton);

    // room code display
    const roomCodeFont = {fontFamily: env.gameFontFamily, color: "#306", fontSize: 128};
    this.roomCodeText = []
    for (let i = 0; i < 4; i++) {
      this.add.rectangle(105 + i * 125, 420, 120, 180).setOrigin(0).setStrokeStyle(1, 0);
      let d = (i < this.roomCode.length) ? this.roomCode[i] : '-';
      this.roomCodeText.push(this.add.text(165 + i * 125, 510, d, roomCodeFont).setOrigin(0.5));
    }

    // check if cookies were set
    let cookieArray = document.cookie.split("; ");
    let cookieRoomCode = cookieArray.find((row) => row.startsWith("room_code="))?.split("=")[1];
    let cookieCardID = cookieArray.find((row) => row.startsWith("card_id="))?.split("=")[1];
    if (cookieRoomCode && cookieCardID) {
      continueRoom.setText("Room Code: " + cookieRoomCode);
      continueCard.setText("Card ID: " + cookieCardID);
      game.roomCode = cookieRoomCode;
      game.cardID = cookieCardID;
      continueButton.enable();
    }
  }

  // (callback) back or number button is clicked
  enterNumber(num) {
    if (num != "<") { // back button
      this.roomCodeText[this.roomCode.length].setText(num);
      this.roomCode += num;
      if (this.roomCode.length == 4) {
        for (let i = 0; i < 10; i++)
          this.numberButton[i].disable();
          this.startButton.enable();
      }
      this.numberButton[10].enable();
    } else { // number button
      this.roomCode = this.roomCode.slice(0, -1);
      this.roomCodeText[this.roomCode.length].setText('-');
      for (let i = 0; i < 10; i++)
        this.numberButton[i].enable();
      this.startButton.disable();
      if (this.roomCode.length == 0)
        this.numberButton[10].disable();
    }
  }

  // (callback) start or continue button is clicked
  checkRoom(arg) {
    if (arg == "Start") { // start button
      game.roomCode = this.roomCode;
      game.cardID = "";
      env.sceneAfterLoad = "cardsetup";
    } else { // continue button
      env.sceneAfterLoad = "main";
    }
    let loader = new RoomInfoLoader(game.roomCode, "read", this.goToGame.bind(this));
  }

  // (callback) room is loaded
  goToGame(info) {
    if (info.valid) {
      this.scene.start("loading");
    } else {
      this.roomStatusText.setText(info.message);
    }
  }
}

/**********************************************************************/
// card setup screen, big cell and its position are determined here
class SetupScreen extends Phaser.Scene {
  constructor( ...args ) {
    super({ key: 'cardsetup', ...args });
  }

  create() {
    this.add.text(640, 50, 'Choose Your Favorite Character', env.defaultFont).setOrigin(0.5, 0);
    this.selectedCell = this.add.existing(new BingoCell(this, 1, 'l', 100, 680));

    // buttons
    let cursorFunc = this.cursorClicked.bind(this);
    let enterFunc = this.enterClicked.bind(this);
    let cancelFunc = this.cancelClicked.bind(this);
    this.add.existing(new GameButton(this, "^", 860, 680, 150, 150, true, cursorFunc));
    this.add.existing(new GameButton(this, "<", 690, 770, 150, 150, true, cursorFunc));
    this.add.existing(new GameButton(this, "v", 860, 860, 150, 150, true, cursorFunc));
    this.add.existing(new GameButton(this, ">", 1030, 770, 150, 150, true, cursorFunc));
    this.add.existing(new GameButton(this, "OK", 960, 1080, 220, 150, false, enterFunc));
    this.add.existing(new GameButton(this, "Back", 690, 1080, 220, 150, false, cancelFunc));

    // icon selector
    this.panel = [];
    for (let i = 0; i < env.numPerColumn; i++) {
      this.panel.push(this.add.plane(640, 400, 'preview_' + (i + 1).toString()));
    }
    this.arrowUp = this.add.triangle(1140, 200, 0, -20, 60, 20, -60, 20, env.columnColorFore['p']);
    this.arrowUp.setStrokeStyle(4, env.columnColorBack['p']);
    this.arrowDown = this.add.triangle(1140, 600, 0, 20, 60, -20, -60, -20, env.columnColorFore['b']);
    this.arrowDown.setStrokeStyle(4, env.columnColorBack['b']);

    // big cell position selector
    let posFrame   = this.add.rectangle(640, 380, 450, 450, 0xc0c0c0).setStrokeStyle(2, 0);
    let posInner   = this.add.rectangle(640, 380, 400, 400, 0xffffff);
    this.posColumn = this.add.rectangle(640, 380, 80, 400, 0xffffff).setOrigin(0, 0.5);
    let posCard    = this.add.rectangle(640, 380, 400, 400).setStrokeStyle(2, 0);
    this.posSelect = this.add.rectangle(640, 380, 160, 160, 0xffffff).setStrokeStyle(2, 0);
    this.posLayer  = this.add.layer([posFrame, posInner, this.posColumn, posCard]).setDepth(1000);
    for (let i = 0; i < 4; i++) {
      this.posLayer.add(this.add.line(520 + i * 80, 380, 0, 0, 0, 400, 0x808080));
      this.posLayer.add(this.add.line(640, 260 + i * 80, 0, 0, 400, 0, 0x808080));
    }
    this.posLayer.add(this.posSelect);
    this.posLayer.setVisible(false);

    // variable initialization
    this.colorIndex = 0;
    this.colorLastIndex = -1;
    this.charIndex = 0;
    this.panelRotate = 180;
    this.panelLastRotate = 180;
    this.panelRotateTarget = 0;
    this.updatePanel();
    this.selectMode = 'chara';
    this.positionR = 0;
    this.positionX = 0;
    this.positionY = 0;
    this.startX = 0;
    this.endX = 0;
  }

  // update icon selector
  updatePanel(deltaFrame = 0) {
    let d = deltaFrame * 0.15;
    this.panelRotate = this.panelRotate * (1 - d) + this.panelRotateTarget * d;
    if (Math.abs(this.panelRotate - this.panelLastRotate) < 0.2)
      return;	
    let colorChanged = (this.colorIndex != this.colorLastIndex);

    for (let i = 0; i < env.numPerColumn; i++) {
      let num = this.colorIndex * env.numPerColumn + i + 1;
      let deg = this.panelRotate + i * 360 / env.numPerColumn;
      let lastDeg = this.panelLastRotate + i * 360 / env.numPerColumn;
      let x = 640 + Math.sin(deg * Math.PI / 180) * 400;
      let y = 380 + Math.cos(deg * Math.PI / 180) * 100;
      let z = 0.8 + Math.cos(deg * Math.PI / 180) * 0.2;
      let normDeg = ((deg % 360) + 360) % 360; // normalize angle to 0-360 deg.
      let normLastDeg = ((lastDeg % 360) + 360) % 360;
      let back = 	(normDeg >= 90 && normDeg <= 270);
      let lastBack = (normLastDeg >= 90 && normLastDeg <= 270);

      this.panel[i].setPosition(x, y);
      this.panel[i].setScale(z);
      this.panel[i].setDepth(y);
      this.panel[i].rotateY = deg;
      if (back && (! lastBack || colorChanged)) {
        this.panel[i].setTexture('preview_b' + env.colors[this.colorIndex]);
      } else if (! back && (lastBack || colorChanged)) {
        this.panel[i].setTexture('preview_' + num.toString());
      }
    }
    if (colorChanged) {
      let u = env.colors[(this.colorIndex + env.colors.length - 1) % env.colors.length];
      this.arrowUp.setFillStyle(env.columnColorFore[u]);
      this.arrowUp.setStrokeStyle(4, env.columnColorBack[u]);
      let d = env.colors[(this.colorIndex + 1) % env.colors.length];
      this.arrowDown.setFillStyle(env.columnColorFore[d]);
      this.arrowDown.setStrokeStyle(4, env.columnColorBack[d]);
    }
    this.colorLastIndex = this.colorIndex;
    this.panelLastRotate = this.panelRotate;
  }

  // update cell position selector
  updatePosSelect() {
    this.posSelect.setPosition(520 + 80 * this.positionX, 260 + 80 * this.positionY);
    this.posSelect.setFillStyle(env.columnColorBack[env.colors[this.colorIndex]]);
    this.posSelect.setAlpha(0.5 + Math.abs(Math.cos(this.positionR)) * 0.5);
  }

  // (callback) cursor button is clicked
  cursorClicked(arg) {
    if (this.selectMode == 'chara') {
      if (arg == '^') {
        this.colorIndex = (this.colorIndex + env.colors.length - 1) % env.colors.length;
        this.panelRotate += 180;
      } else if (arg == 'v') {
        this.colorIndex = (this.colorIndex + 1) % env.colors.length;
        this.panelRotate += 180;
      } else if (arg == '<') {
        this.charIndex = (this.charIndex + env.numPerColumn - 1) % env.numPerColumn;
        this.panelRotateTarget += 360 / env.numPerColumn;
      } else if (arg == '>') {
        this.charIndex = (this.charIndex + 1) % env.numPerColumn;
        this.panelRotateTarget -= 360 / env.numPerColumn;
      }
      this.selectedCell.setNum(this.charIndex + this.colorIndex * env.numPerColumn + 1);
    } else if (this.selectMode == 'position') {
      if (arg == '^') {
        this.positionY = Math.max(this.positionY - 1, 0);
      } else if (arg == 'v') {
        this.positionY = Math.min(this.positionY + 1, 3);
      } else if (arg == '<') {
        this.positionX = Math.max(this.positionX - 1, this.startX);
      } else if (arg == '>') {
        this.positionX = Math.min(this.positionX + 1, this.endX);
      }
    }
  }

  // (callback) OK button is clicked
  enterClicked(arg) {
    if (this.selectMode == 'chara') {
      // initialize position selector
      this.startX = Math.max(0, this.colorIndex - 1); 
      this.endX   = Math.min(env.colors.length - 2, this.colorIndex);
      this.posColumn.setDisplaySize(80 * (this.endX - this.startX + 2), 400);
      this.posColumn.setPosition(440 + 80 * this.startX, 380);
      this.posColumn.setFillStyle(env.columnColorFore[env.colors[this.colorIndex]]);
      this.positionX = this.startX;
      this.positionY = 0;
      this.positionR = 0;
      this.updatePosSelect();
      this.posLayer.setVisible(true);
      this.selectMode = 'position';
    } else if (this.selectMode == 'position') {
      // generate card ID and proceed to the main screen
      game.cardID = gen.decToB26(this.positionX + this.positionY * 5);
      let idx = this.charIndex + (this.colorIndex - this.positionX) * env.numPerColumn;
      game.cardID += gen.decToB26(idx);
      let rnd = new Uint32Array(1);
      crypto.getRandomValues(rnd);
      game.cardID += gen.decToB26(rnd[0]).padStart(7, 'A');
      document.cookie = "room_code=" + game.roomCode + "; SameSite=Strict; max-age=10000";
      document.cookie = "card_id=" + game.cardID + "; SameSite=Strict; max-age=10000";
      this.scene.start("main");
    }
  }

  // (callback) cancel button is clicked
  cancelClicked(arg) {
    if (this.selectMode == 'chara') {
      this.scene.start("title");
    } else if (this.selectMode == 'position') {
      this.posLayer.setVisible(false);
      this.selectMode = 'chara';
      this.panelRotate += 180;
    }
  }

  update(time, delta) {
    let deltaFrame = Math.min(delta * 0.03, 1);
    if (this.selectMode == 'chara') {
      this.updatePanel(deltaFrame);
    } else if (this.selectMode == 'position') {
      this.positionR += deltaFrame * 0.05;
      this.updatePosSelect();
    }
  }
}

/**********************************************************************/
// main game screen, display the bingo card and check lines
class MainScreen extends Phaser.Scene {
  constructor( ...args ) {
    super({ key: 'main', ...args });
  }

  create() {
    let cols = env.colors.length;
    this.card = gen.generate(game.cardID);

    // cells
    this.cells = [];
    for (let i = 0; i < this.card.length; i++) {
      if (this.card[i] > 0) {
        let x = 20 + (i % cols) * 250;
        let y = 20 + Math.floor(i / cols) * 250;
        let s = (i < this.card.length - cols - 1 && this.card[i + cols + 1] == -cols - 1) ? 'l' :
                (i < this.card.length - cols     && this.card[i + cols]     == -cols    ) ? 'v' :
                (i < this.card.length - 1        && this.card[i + 1]        == -1       ) ? 'h' : 's';
        this.cells.push(this.add.existing(new BingoCell(this, this.card[i], s, x, y)));
        if (s == 'l') {
          this.mainColor = env.colors[Math.floor((this.card[i] - 1) / env.numPerColumn)];
        }
      } else {
        this.cells.push(this.cells[i + this.card[i]]);
      }
    }
    this.cameras.main.setBackgroundColor(env.columnColorBack[this.mainColor]);

    // lines (horizontal and vertical)
    this.lines = [];
    for (let i = 0; i < cols; i++) {
      let hLine = {cells: [], indicators: [], tween: null, rest: -1};
      let vLine = {cells: [], indicators: [], tween: null, rest: -1};
      for (let j = 0; j < cols; j++) {
        if (this.card[i * cols + j] % cols != -1)
          hLine.cells.push(this.cells[i * cols + j]);
        if (this.card[i + j * cols] > -cols)
          vLine.cells.push(this.cells[i + j * cols]);
      }
      hLine.indicators.push(this.add.rectangle(8, 140 + i * 250, 15, 120, 0xffffff).setStrokeStyle(2, 0));
      hLine.indicators.push(this.add.rectangle(1272, 140 + i * 250, 15, 120, 0xffffff).setStrokeStyle(2, 0));
      vLine.indicators.push(this.add.rectangle(140 + i * 250, 8, 120, 15, 0xffffff).setStrokeStyle(2, 0));
      vLine.indicators.push(this.add.rectangle(140 + i * 250, 1272, 120, 15, 0xffffff).setStrokeStyle(2, 0));
      this.lines.push(hLine);
      this.lines.push(vLine);
    }

    // lines (diagonal)
    let dLine1 = {cells: [], indicators: [], tween: null, rest: -1};
    let dLine2 = {cells: [], indicators: [], tween: null, rest: -1};
    for (let i = 0; i < cols; i++) {
      if (this.card[i * (cols + 1)] != -cols - 1)
        dLine1.cells.push(this.cells[i * (cols + 1)]);
      if (this.card[i * (cols - 1)] != -1 || this.card[(i + 1) * (cols - 1)] != -cols)
        dLine2.cells.push(this.cells[(i + 1) * (cols - 1)]);
    }
    dLine1.indicators.push(this.add.rectangle(20, 20, 38, 38, 0xffffff).setStrokeStyle(2, 0));
    dLine1.indicators.push(this.add.rectangle(1260, 1260, 38, 38, 0xffffff).setStrokeStyle(2, 0));
    dLine2.indicators.push(this.add.rectangle(20, 1260, 38, 38, 0xffffff).setStrokeStyle(2, 0));
    dLine2.indicators.push(this.add.rectangle(1260, 20, 38, 38, 0xffffff).setStrokeStyle(2, 0));
    this.lines.push(dLine1);
    this.lines.push(dLine2);

    // layer on error
    let confirmFunc = this.confirmClicked.bind(this);
    let conFrame = this.add.rectangle(1040, 130, 400, 180, 0xffffff).setStrokeStyle(2, 0);
    this.statusText = this.add.text(1040, 60, "Status", env.defaultFont).setOrigin(0.5, 0);
    let conRetryButton = new GameButton(this, "Retry", 870, 130, 160, 70, false, confirmFunc);
    let conTitleButton = new GameButton(this, "Title", 1050, 130, 160, 70, false, confirmFunc);
    this.confirmLayer = this.add.layer([conFrame, this.statusText, conRetryButton, conTitleButton]);
    this.confirmLayer.setDepth(10);
    this.confirmLayer.setVisible(false);

    this.lastSecond = 0;
    this.waitSecond = 0;
    this.lastDrawnLength = 0;
    this.drawn = [];
  }

  // update tweens for line indicators
  updateLineTweens() {
    for (let i = 0; i < this.lines.length; i++) {
      let rest = this.lines[i].cells.map((x) => (x.hit) ? 0 : 1). reduce((s, x) => s + x);
      if (rest == this.lines[i].rest)
        continue;
      
      let newCounter = {duration: 0, yoyo: true, loop: -1, onUpdate: null};
      this.lines[i].rest = rest;
      if (this.lines[i].tween)
        this.lines[i].tween.stop();
      if (rest == 2) {
        newCounter.duration = 1500;
        newCounter.onUpdate = (t) => { this.onNearReach(this.lines[i].indicators, t.getValue()); };
      } else if (rest == 1) {
        newCounter.duration = 750;
        newCounter.onUpdate = (t) => { this.onReach(this.lines[i].indicators, t.getValue()); };
      } else if (rest == 0) {
        newCounter.duration = 1000;
        newCounter.yoyo = false;
        newCounter.onUpdate = (t) => { this.onBingo(this.lines[i].indicators, t.getValue()); };
      }
      this.lines[i].tween = newCounter.onUpdate && this.tweens.addCounter(newCounter);
    }
  }

  // (callback) set indicator color for near-reach line
  onNearReach(targets, value) {
    let r = Math.floor(value * value * 255);
    let g = Math.floor(128 + r / 2);
    let c = r * 0x10000 + g * 0x100 + 0xff;
    for (let i = 0; i < targets.length; i++)
      targets[i].setFillStyle(c);
  }

  // (callback) set indicator color for reach line
  onReach(targets, value) {
    let b = Math.floor(value * value * 255);
    let c = 0xffff00 + b;
    for (let i = 0; i < targets.length; i++)
      targets[i].setFillStyle(c);
  }

  // (callback) set indicator color for bingo line
  onBingo(targets, value) {
    let c = env.rainbow(Math.min(0.999, value)); // ensure value < 1.0
    for (let i = 0; i < targets.length; i++)
      targets[i].setFillStyle(c);
  }

  // (callback) new room data is loaded
  roomLoadFinished(info) {
    if (info.valid) {
      this.drawn = info.drawn;
      this.waitSecond = 10;
    } else {
      this.statusText.setText(info.message);
      this.confirmLayer.setVisible(true);
    }
  }

  // (callback) retry or title button is clicked
  confirmClicked(arg) {
    if (arg == "Retry") {
      this.confirmLayer.setVisible(false);
      this.waitSecond = 0;
    } else {
      this.scene.start("title");
    }
  }

  update() {
    let tm = Math.floor(Date.now() / 1000);
    if (tm == this.lastSecond || this.waitSecond == -1)
      return;
    this.lastSecond = tm;
    
    // timer for room data loading
    if (this.waitSecond == 0) {
      let loader = new RoomInfoLoader(game.roomCode, "read", this.roomLoadFinished.bind(this));
    }
    this.waitSecond--;
    if (this.drawn.length == this.lastDrawnLength)
      return;

    // update cells
    let roomUpdated = false;
    for (let i = this.lastDrawnLength; i < this.drawn.length; i++) {
      let diffTm = tm - this.drawn[i].tm;
      if (diffTm < 0)
        continue;

      this.lastDrawnLength = i + 1;
      let f = this.cells.find((x) => x.nums.includes(this.drawn[i].num));
      if (!f)
        continue;

      roomUpdated = true;
      f.hit = true;
      if (diffTm <= 20) {
        this.tweens.addCounter({
          duration: 1000,
          loop: 10,
          onUpdate: (t) => this.onBingo([f.base], t.getValue()),
          onComplete: (t) => f.setBaseColor(0xff0040)});
      } else {
        f.setBaseColor(0xff0040);
      }
    }
    if (roomUpdated)
      this.updateLineTweens();
  }
}

/**********************************************************************/
const config = {
  type: Phaser.AUTO,
  scale: {
    mode: Phaser.Scale.FIT,
    autoCenter: Phaser.Scale.CENTER_HORIZONTALLY},
  width: 1280,
  height: 1280,
  backgroundColor: '#fff',
  parent: 'game-container',
  scene: [TitleScreen, LoadScreen, SetupScreen, MainScreen],
  fps: {
      target: 30
  }
};

/**********************************************************************/