// bingo2 for PLB (C) 2024 Shimeji Sola
// A bingo lottery client played by the game manager

// Global variables of the client, shared among the screens
class ClientEnvironment {
  constructor() {
    this.roomCode = "";
    this.cardID = "";
  }
}

var game = new ClientEnvironment();

/**********************************************************************/
// title screen of the client, room code is determined here
class TitleScreen extends Phaser.Scene {
  constructor( ...args ) {
    super({ key: 'title', ...args });
  }

  create() {
    this.roomCode = "";

    // texts
    this.add.text(400, 150, 'Bingo 2 for PLB - Lottery', env.defaultFont).setOrigin(0.5, 0);
    this.add.text(400, 250, 'Room Code', env.defaultFont).setOrigin(0.5, 0);
    this.roomStatusText = this.add.text(400, 800, '', env.defaultFont).setOrigin(0.5, 0.5);

    // buttons
    let startFunc = this.checkRoom.bind(this);
    let enterNumberFunc = this.enterNumber.bind(this);
    this.startButton = new GameButton(this, "Start", 150, 900, 500, 200, false, startFunc);
    this.numberButton = [];
    for (let y = 0; y < 4; y++) {
      for (let x = 0; x < 3; x++) {
        if (y != 3 || x != 0) {
          let n = "123456789 0<"[x + y * 3];
          let newButton = new GameButton(this, n, 800 + x * 250, 150 + y * 250, 200, 200, true, enterNumberFunc);
          this.numberButton.push(newButton);
          this.add.existing(newButton);
        }
      }
    }
    this.startButton.disable();
    this.numberButton[10].disable();
    this.add.existing(this.startButton);

    // room code display
    const roomCodeFont = {fontFamily: env.gameFontFamily, color: "#306", fontSize: 144};
    this.roomCodeText = []
    for (let i = 0; i < 4; i++) {
      this.add.rectangle(105 + i * 145, 450, 140, 250).setOrigin(0).setStrokeStyle(1, 0);
      let d = (i < this.roomCode.length) ? this.roomCode[i] : '-';
      this.roomCodeText.push(this.add.text(175 + i * 145, 575, d, roomCodeFont).setOrigin(0.5));
    }
  }

  // (callback) back or number button is clicked
  enterNumber(num) {
    if (num != "<") { // back button
      this.roomCodeText[this.roomCode.length].setText(num);
      this.roomCode += num;
      if (this.roomCode.length == 4) {
        for (let i = 0; i < 10; i++)
          this.numberButton[i].disable();
          this.startButton.enable();
      }
      this.numberButton[10].enable();
    } else { // number button
      this.roomCode = this.roomCode.slice(0, -1);
      this.roomCodeText[this.roomCode.length].setText('-');
      for (let i = 0; i < 10; i++)
        this.numberButton[i].enable();
      this.startButton.disable();
      if (this.roomCode.length == 0)
        this.numberButton[10].disable();
    }
  }

  // (callback) start button is clicked
  checkRoom(arg) {
    game.roomCode = this.roomCode;
    env.sceneAfterLoad = "main";
    let loader = new RoomInfoLoader(game.roomCode, "open", this.goToGame.bind(this));
  }

  // (callback) room is loaded
  goToGame(info) {
    if (info.valid) {
      this.scene.start("loading");
    } else {
      this.roomStatusText.setText(info.message);
    }
  }
}

/**********************************************************************/
// main game screen, display the bingo card and check lines
class MainScreen extends Phaser.Scene {
  constructor( ...args ) {
    super({ key: 'main', ...args });
  }

  create() {
    // number list
    this.numbers = [];
    let w = env.colors.length;
    let h = env.numPerColumn;
    for (let i = 0; i < w; i++) {
      for (let j = 0; j < h; j++) {
        let n = i * env.numPerColumn + j + 1;
        let c = new BingoCell(this, n, 's', i * 640 / w, j * 1200 / h, 5 / h);
        c.setIconAlpha(0.3);
        this.numbers.push(this.add.existing(c));
      }
    }
    
    // animation area
    let animRect = this.add.rectangle(1120, 480, 960, 960, 0xffffff).setDepth(2).setAlpha(0.5);
    let animFrame = this.add.rectangle(1120, 480, 958, 958).setStrokeStyle(2, 0).setDepth(2);
    this.animMask = animRect.createGeometryMask();
    this.backIcons = [];
    this.backBoost = 0;

    // confirm window
    let conFrame = this.add.rectangle(800, 600, 900, 300, 0xffffff).setStrokeStyle(2, 0);
    let conText = this.add.text(800, 520, "Are you okay with closing this room?", env.defaultFont)
    conText.setOrigin(0.5);

    // load the card list
    let loader = new RoomInfoLoader(game.roomCode, "read", this.firstLoadFinished.bind(this));

    // buttons
    let drawFunc = this.drawClicked.bind(this);
    let closeFunc = this.closeClicked.bind(this);
    let confirmFunc = this.confirmClicked.bind(this);
    this.drawButton = new GameButton(this, "Draw", 740, 990, 320, 140, false, drawFunc);
    this.closeButton = new GameButton(this, "Close", 1160, 990, 320, 140, false, closeFunc);
    this.statusText = this.add.text(1110, 1130, "", env.defaultFont).setOrigin(0.5, 0);
    let conYesButton = new GameButton(this, "Yes", 540, 590, 240, 120, false, confirmFunc);
    let conNoButton = new GameButton(this, "No", 820, 590, 240, 120, false, confirmFunc);
    this.add.existing(this.drawButton);
    this.add.existing(this.closeButton);
    this.confirmLayer = this.add.layer([conFrame, conText, conYesButton, conNoButton]);
    this.confirmLayer.setDepth(10);
    this.confirmLayer.setVisible(false);

    // variables
    this.drawFrame = null;
    this.drawCards = [];
    this.drawNum = 0;
    this.lotoIndex = 0;
    this.restNums = [];
  }

  // update background icons
  updateBackIcons(deltaFrame) {
    // move down existing icons
    for (let i = 0; i < this.backIcons.length; i++) {
      let deg = this.backIcons[i].obj.rotateY;
      deg = (deg + deltaFrame * 2) % 360;
      this.backIcons[i].obj.rotateY = deg;
      this.backIcons[i].obj.y += deltaFrame * 5;
      if (deg >= 90 && deg < 92) {
        this.backIcons[i].obj.setTexture('preview_b' + this.backIcons[i].col);
      } else if (deg >= 270 && deg < 272) {
        this.backIcons[i].obj.setTexture('preview_' + this.backIcons[i].num.toString());
      }
      if (this.backIcons[i].obj.y > 1200) {
        this.backIcons[i].fin = true;
        this.backIcons[i].obj.destroy();
      }
    }
    this.backIcons = this.backIcons.filter((x) => ! x.fin);

    // generate new icon
    if (this.backIcons.length < 20 && Math.random() < 0.05 * deltaFrame * this.backBoost) {
      let newNum = Math.floor(Math.random() * env.numberImages.length) + 1;
      let newCol = env.colors[Math.floor((newNum - 1) / env.numPerColumn)];
      let newObj = this.add.plane(640 + Math.random() * 960, -240, 'preview_' + newNum.toString());
      let scale = 0.3 + Math.random() * 0.5;
      newObj.setScale(scale);
      newObj.setDepth(scale);
      newObj.rotateX = -45 + Math.random() * 90;
      newObj.rotateY = -90 + Math.random() * 180;
      newObj.rotateZ = -45 + Math.random() * 90;
      newObj.setMask(this.animMask);
      this.backIcons.push({obj: newObj, num: newNum, col: newCol, fin: false});
      this.backBoost = 0
    } else {
      this.backBoost += 0.1 * deltaFrame;
    }
  }

  // initialize lottery animation (cells and lamp in number list)
  initiateDrawAnimation(num) {
    let col = env.colors[Math.floor((num - 1) / env.numPerColumn)];
    this.drawFrame = 300;
    this.drawNum = num; 
    this.restNums = this.numbers.filter((x) => ! x.hit).map((x) => x.nums[0]);
    this.lotoIndex = this.restNums.length - 1;
    for (let i = 0; i < 17; i++) {
      let newNum = (i == 0) ? num : this.restNums[Math.floor(Math.random() * this.restNums.length)];
      let newCell = new BingoCell(this, newNum, 'l', 875, -740 * i + 235, 1, false);
      newCell.setMask(this.animMask);
      newCell.setDepth(3);
      this.drawCards.push(this.add.existing(newCell));
    }
  }

  // update lottery animation
  updateDrawAnimation(deltaFrame) {
    let lastFrame = this.drawFrame;
    this.drawFrame -= deltaFrame;

    // cells
    let offset = (this.drawFrame >=  150) ? this.drawFrame * 60 - 5400 :
                 (this.drawFrame >=    0) ? Math.pow((this.drawFrame - 15), 2) / 5 - 45 :
                 (this.drawFrame <= -120) ? -Math.pow((this.drawFrame + 120), 2) : 0;
    for (let i = 0; i < this.drawCards.length; i++)
      this.drawCards[i].setOffset(0, offset);

    // number list lamp
    if (Math.floor(lastFrame) != Math.floor(this.drawFrame)) {
      if (this.drawFrame >= 15 || this.drawNum != this.restNums[this.lotoIndex]) {
        let index = this.restNums[this.lotoIndex] - 1;
        this.numbers[index].setBaseColor(0xffffff);
        this.lotoIndex = (this.lotoIndex + 1) % this.restNums.length;
        index = this.restNums[this.lotoIndex] - 1;
        this.numbers[index].setBaseColor(0x4400ff);
      } else if (this.drawFrame >= -140) {
        let v = ((15 - this.drawFrame) % 30) / 30;
        this.numbers[this.drawNum - 1].setBaseColor(env.rainbow(v));
        this.numbers[this.drawNum - 1].setIconAlpha(1);
        this.numbers[this.drawNum - 1].hit = true;
      } else {
        this.numbers[this.drawNum - 1].setBaseColor(0xff0044);
      }
    }

    // end of animation
    if (this.drawFrame <= -150) {
      for (let i = 0; i < this.drawCards.length; i++)
        this.drawCards[i].destroy();
      this.drawFrame = null;
      this.drawCards = [];
      this.drawButton.enable();
      this.closeButton.enable();
    }
  }

  // (callback) draw button is clicked
  drawClicked(arg) {
    this.drawButton.disable();
    this.closeButton.disable();
    let loader = new RoomInfoLoader(game.roomCode, "draw", this.drawRequestFinished.bind(this));
  }

  // (callback) close button is clicked
  closeClicked(arg) {
    this.drawButton.disable();
    this.closeButton.disable();
    this.confirmLayer.setVisible(true);
  }
  
  // (callback) confirm button (yes/no) is clicked
  confirmClicked(arg) {
    if (arg == "Yes") {
      this.confirmLayer.setVisible(false);
      let loader = new RoomInfoLoader(game.roomCode, "close", this.closeRequestFinished.bind(this));
    } else {
      this.drawButton.enable();
      this.closeButton.enable();
      this.confirmLayer.setVisible(false);
    }
  }

  // (callback) finished first load
  firstLoadFinished(info) {
    if (! info.valid) {
      this.statusText.setText(info.message);
      return;
    }
    this.refreshCards(info.drawn, info.drawn.length);
  }

  // (callback) finished draw request
  drawRequestFinished(info) {
    if (! info.valid) {
      this.statusText.setText(info.message);
      this.drawButton.enable();
      this.closeButton.enable();
      return;
    }
    this.refreshCards(info.drawn, info.drawn.length - 1);
    this.initiateDrawAnimation(info.drawn[info.drawn.length - 1].num);
  }

  // (callback) finished close request
  closeRequestFinished(info) {
    if (info.message != "Room Closed") {
      this.statusText.setText(info.message);
      this.drawButton.enable();
      this.closeButton.enable();
      return;
    }
    this.scene.start("title");
  }

  // common routine for refreshing the list of drawn cards
  refreshCards(drawn, endIndex) {
    let nums = drawn.slice(0, endIndex).map((x) => x.num);
    for (let i = 0; i < this.numbers.length; i++) {
      if (nums.includes(i + 1)) {
        this.numbers[i].setBaseColor(0xff0044);
        this.numbers[i].setIconAlpha(1);
        this.numbers[i].hit = true;
      } else {
        this.numbers[i].setBaseColor(0xffffff);
        this.numbers[i].setIconAlpha(0.3);
        this.numbers[i].hit = false;
      }
    }
  }

  update(time, delta) {
    let deltaFrame = Math.min(delta * 0.03, 1);
    this.updateBackIcons(deltaFrame);
    if (this.drawFrame)
      this.updateDrawAnimation(deltaFrame);
  }
}

/**********************************************************************/
const config = {
  type: Phaser.AUTO,
  scale: {
    mode: Phaser.Scale.FIT,
    autoCenter: Phaser.Scale.CENTER_HORIZONTALLY},
  width: 1600,
  height: 1200,
  backgroundColor: '#fff',
  parent: 'game-container',
  scene: [TitleScreen, LoadScreen, MainScreen],
  fps: {
      target: 30
  }
};

/**********************************************************************/